"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var tslib_1 = require("tslib");
var constants_1 = require("./constants");
var errors_1 = tslib_1.__importDefault(require("./errors"));
var utils = tslib_1.__importStar(require("./utils"));
var SchemaNormalizer = /** @class */ (function () {
    function SchemaNormalizer(schema) {
        this.schema = schema;
        this.db = schema.db;
    }
    SchemaNormalizer.prototype.normalize = function (data, context) {
        var _this = this;
        if (!utils.isPlainObject(data) && !Array.isArray(data))
            throw new Error(errors_1.default.normalizeInvalidData());
        utils.ensureParamObject("context", context);
        utils.ensureParamObject("context.output", context.output);
        if (!context.output[this.schema.name])
            context.output[this.schema.name] = constants_1.initialState(this.schema.name);
        return utils.ensureArray(data).map(function (obj) {
            if (!utils.isPlainObject(obj))
                throw new Error(errors_1.default.normalizeInvalidData());
            var subject = obj;
            var normalizer = _this.db.getRecordNormalizer(_this.schema.name);
            if (normalizer)
                subject = normalizer(subject, context);
            var pk = context.normalizePKs
                ? _this._normalizePrimaryKey(subject)
                : _this.schema.getPrimaryKey(subject);
            if (!pk)
                throw new Error(errors_1.default.normalizePk(_this.schema.name));
            var tbl = context.output[_this.schema.name];
            if (!tbl.byId[pk])
                tbl.ids.push(pk);
            var record = tbl.byId[pk] = tslib_1.__assign({}, subject);
            var fks = _this.schema.getForeignKeys(subject);
            fks.forEach(function (fk) {
                _this._normalizeForeignKey(fk, record, context);
                _this._indexForeignKey(fk, tbl, pk);
            });
            _this.schema.relations.forEach(function (rel) {
                return _this._normalizeRelations(rel, record, pk, context);
            });
            return pk;
        });
    };
    SchemaNormalizer.prototype._normalizePrimaryKey = function (record) {
        var pk = this.schema.getPrimaryKey(record);
        if (pk)
            return pk;
        // Invoke the "onGeneratePK" hook if PK not found.
        var generator = this.db.getPkGenerator(this.schema.name);
        if (!generator)
            return undefined;
        var generatedPk = generator(record, this.schema);
        if (generatedPk)
            // if the PK is generated and we have a single PK field definition, then inject it into the record.
            if (this.schema.primaryKeys.length === 1)
                record[this.schema.primaryKeys[0].propName] = generatedPk;
        // Handling multiple PK field defs:
        // We may need the "onGeneratePK" hook to return an object defining each key value.
        // BUT this seems like a rare scenario..
        // So for now; don't populate record.
        return generatedPk;
    };
    SchemaNormalizer.prototype._normalizeRelations = function (rel, record, pk, context) {
        // Normalize foreign relations, FKs from other tables referencing this table.
        // Then remove the nested relations from the record.
        if (rel.relationName && record[rel.relationName]) {
            var normalizedRels = this.schema.inferRelations(record[rel.relationName], rel, pk);
            rel.table.normalize(normalizedRels, context);
            delete record[rel.relationName];
        }
    };
    SchemaNormalizer.prototype._normalizeForeignKey = function (fk, record, context) {
        // if the FK is an object, then normalize it and replace object with it's PK.
        if (typeof fk.value !== "object")
            return;
        var fkPks = fk.refTable.normalize(fk.value, context);
        if (fkPks.length > 1)
            throw new Error(errors_1.default.normalizeInvalidFk(this.schema.name, fk.name, fk.refTable.name));
        record[fk.name] = fk.value = fkPks[0];
    };
    SchemaNormalizer.prototype._indexForeignKey = function (fk, tbl, pk) {
        if (!utils.isValidID(fk.value))
            return;
        var fkId = utils.asID(fk.value);
        var idx = tbl.indexes[fk.name]
            || (tbl.indexes[fk.name] = { unique: fk.unique, values: {} });
        if (!idx.values[fkId])
            idx.values[fkId] = [];
        if (idx.unique && idx.values.length)
            throw new Error(errors_1.default.normalizeFkViolation(this.schema.name, fk.name));
        idx.values[fkId].push(pk);
    };
    return SchemaNormalizer;
}());
exports.default = SchemaNormalizer;
