"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var tslib_1 = require("tslib");
var constants_1 = require("../constants");
var errors_1 = tslib_1.__importDefault(require("../errors"));
var state_1 = tslib_1.__importDefault(require("../state"));
var utils = tslib_1.__importStar(require("../utils"));
var NormalizeContext_1 = tslib_1.__importDefault(require("./NormalizeContext"));
var TableModel = /** @class */ (function () {
    function TableModel(session, schema, state) {
        this.dirty = false;
        this.session = utils.ensureParamObject("session", session);
        this.schema = utils.ensureParamObject("schema", schema);
        this.state = utils.ensureParamObject("state", state || constants_1.initialState(this.schema.name));
        var _a = this.state, ids = _a.ids, byId = _a.byId, indexes = _a.indexes, name = _a.name;
        if (!ids || !byId || !indexes)
            throw new Error(errors_1.default.tableInvalidState(schema.name));
        if (!name)
            this.state.name = name;
    }
    Object.defineProperty(TableModel.prototype, "length", {
        get: function () {
            return this.state.ids.length;
        },
        enumerable: true,
        configurable: true
    });
    TableModel.prototype.all = function () {
        var _this = this;
        return this.state.ids.map(function (id) { return _this.schema.db.factory.newRecordModel(id, _this); });
    };
    TableModel.prototype.values = function () {
        var _this = this;
        return this.state.ids.map(function (id) { return _this.state.byId[id]; });
    };
    TableModel.prototype.exists = function (id) {
        if (!utils.isValidID(id))
            return false;
        return this.state.byId[utils.asID(id)] !== undefined;
    };
    TableModel.prototype.get = function (id) {
        if (!this.exists(id))
            throw new Error(errors_1.default.recordNotFound(this.schema.name, id));
        return this.schema.db.factory.newRecordModel(utils.asID(id), this);
    };
    TableModel.prototype.getOrDefault = function (id) {
        if (!this.exists(id))
            return undefined;
        return this.schema.db.factory.newRecordModel(utils.asID(id), this);
    };
    TableModel.prototype.getValue = function (id) {
        if (utils.isValidID(id))
            return this.state.byId[id];
        else
            return undefined;
    };
    TableModel.prototype.getIndex = function (name, fk) {
        utils.ensureParamString("name", name);
        utils.ensureParamString("fk", fk);
        if (this.state.indexes[name] && this.state.indexes[name].values[fk])
            return this.state.indexes[name].values[fk];
        else
            return [];
    };
    TableModel.prototype.insert = function (data, argument) {
        return this._normalizedAction(data, this.insertNormalized, { normalizePKs: true, argument: argument });
    };
    TableModel.prototype.update = function (data, argument) {
        return this._normalizedAction(data, this.updateNormalized, { argument: argument });
    };
    TableModel.prototype.upsert = function (data, argument) {
        return this._normalizedAction(data, this.upsertNormalized, { normalizePKs: true, argument: argument });
    };
    TableModel.prototype.delete = function (data) {
        var _this = this;
        utils.ensureParam("data", data);
        var idsToDelete = utils.ensureArray(data).map(function (subject) {
            return utils.isObject(subject)
                ? _this.schema.ensurePrimaryKey(subject)
                : utils.ensureID(subject);
        });
        if (!idsToDelete.length)
            return 0;
        this._deleteCascade(idsToDelete);
        var _a = state_1.default.splice(this.state, idsToDelete), deleted = _a.deleted, state = _a.state;
        deleted.forEach(function (_a) {
            var id = _a.id, record = _a.record;
            return _this._cleanIndexes(id, record, state.indexes);
        });
        this.dirty = true;
        this.state = state;
        return deleted.length;
    };
    TableModel.prototype.deleteAll = function () {
        if (this.length) {
            this._deleteCascade(this.state.ids);
            this.state = constants_1.initialState(this.schema.name);
        }
    };
    TableModel.prototype.upsertNormalized = function (norm) {
        var _this = this;
        utils.ensureParamObject("table", norm);
        var toUpdate = constants_1.initialState(this.schema.name);
        var toInsert = constants_1.initialState(this.schema.name);
        norm.ids.forEach(function (id) {
            if (_this.exists(id)) {
                toUpdate.ids.push(id);
                toUpdate.byId[id] = norm.byId[id];
            }
            else {
                toInsert.ids.push(id);
                toInsert.byId[id] = norm.byId[id];
            }
        });
        if (toInsert.ids.length)
            this.insertNormalized(toInsert, false);
        if (toUpdate.ids.length)
            this.updateNormalized(toUpdate, false);
        this._updateIndexes(norm);
    };
    TableModel.prototype.insertNormalized = function (table, updateIndexes) {
        if (updateIndexes === void 0) { updateIndexes = true; }
        utils.ensureParamObject("table", table);
        this.state = state_1.default.merge(this.state, table);
        this.dirty = true;
        if (updateIndexes)
            this._updateIndexes(table);
    };
    TableModel.prototype.updateNormalized = function (table, updateIndexes) {
        var _this = this;
        if (updateIndexes === void 0) { updateIndexes = true; }
        utils.ensureParamObject("table", table);
        var byId = table.ids.reduce(function (map, id) {
            var _a;
            if (!_this.exists(id))
                throw new Error(errors_1.default.recordUpdateViolation(_this.schema.name, id));
            var oldRecord = _this.state.byId[id];
            var newRecord = _this.schema.mergeRecord(oldRecord, table.byId[id]);
            var isModified = _this.schema.isModified(oldRecord, newRecord);
            if (isModified)
                if (map)
                    map[id] = newRecord;
                else
                    return _a = {}, _a[id] = newRecord, _a;
            return map;
        }, null);
        if (byId) {
            this.dirty = true;
            this.state = state_1.default.merge(this.state, { byId: byId });
            if (updateIndexes)
                this._updateIndexes(table);
        }
    };
    TableModel.prototype.normalize = function (data, options) {
        utils.ensureParam("data", data);
        var ctx = new NormalizeContext_1.default(this.schema, tslib_1.__assign({ table: this, normalizePKs: true }, options));
        this.schema.normalize(data, ctx);
        return ctx;
    };
    TableModel.prototype._normalizedAction = function (data, action, options) {
        utils.ensureParam("data", data);
        utils.ensureParamFunction("action", action);
        var ctx = this.normalize(data, options);
        var table = ctx.output[this.schema.name];
        if (table)
            action.call(this, table);
        this.session.upsert(ctx);
        return table.ids;
    };
    TableModel.prototype._cleanIndexes = function (id, record, indexes) {
        var fks = this.schema.getForeignKeys(record);
        state_1.default.cleanIndexes(fks, id, indexes);
    };
    TableModel.prototype._updateIndexes = function (table) {
        state_1.default.updateIndexes(this.schema.name, this.state, table);
    };
    TableModel.prototype._deleteCascade = function (ids) {
        var _this = this;
        var cascade = this.schema.relations.filter(function (rel) { return rel.relationName && rel.cascade; });
        if (cascade.length)
            ids.forEach(function (id) {
                var model = _this.get(id);
                if (model)
                    cascade.forEach(function (schema) {
                        var relation = schema.relationName && model[schema.relationName];
                        if (relation)
                            relation.delete();
                    });
            });
    };
    return TableModel;
}());
exports.default = TableModel;
