export declare type FieldType = "ATTR" | "MODIFIED" | "PK";
export declare type ValueType<R> = R extends {
    value: infer ValueType;
} ? ValueType : never;
export declare type Values<R> = ValueType<R> | ValueType<R>[];
export declare type PartialValue<R> = Partial<ValueType<R>>;
export declare type PartialValues<R> = PartialValue<R> | (PartialValue<R>[]);
export interface RecordEntity {
    id: string;
}
export declare type Reducer = (session: any, action: any, arg?: any) => void;
export declare type PkGenerator = (record: any, schema: TableSchema) => string | null | undefined;
export declare type RecordNormalizer = (record: any, context: NormalizeContext) => any;
export declare type RecordComparer = (a: any, b: any, schema: TableSchema) => boolean;
export declare type RecordMerger = (a: any, b: any, schema: TableSchema) => any;
export interface MapOf<T> {
    [key: string]: T;
}
export interface ForeignKey {
    name: string;
    value: string;
    references: string;
    refTable: TableSchema;
    unique: boolean;
    notNull: boolean;
}
export interface TableSchema {
    db: DatabaseSchema;
    name: string;
    fields: FieldSchema[];
    relations: FieldSchema[];
    primaryKeys: FieldSchema[];
    connect(schemas: MapOf<TableSchema>): void;
    ensurePrimaryKey(record: any): string;
    getPrimaryKey(record: any): string | undefined;
    getForeignKeys(record: any): ForeignKey[];
    isModified(recordA: any, recordB: any): boolean;
    inferRelations(data: any, rel: FieldSchema, ownerId: string): any[];
    injectKeys(data: any, record: TableRecord): void;
    normalize(data: any, context: NormalizeContext): string[];
    mergeRecord(oldRecord: any, newRecord: any): any;
}
export interface FieldSchema {
    table: TableSchema;
    refTable?: TableSchema;
    name: string;
    propName: string;
    references?: string;
    relationName?: string;
    cascade?: boolean;
    unique?: boolean;
    isPrimaryKey: boolean;
    isForeignKey: boolean;
    isStamp: boolean;
    notNull: boolean;
    connect(schemas: MapOf<TableSchema>): void;
    getValue(data: any, record?: any): any;
    getRecordValue(record: any): any;
}
/**
 * Represents an immutable wrapper for a db table.
 *
 * @export
 * @interface Table
 * @template R
 */
export interface Table<R extends TableRecord = TableRecord> {
    session: Session;
    schema: TableSchema;
    state: TableState;
    dirty: boolean;
    /**
     * Gets a single record by it's PK.
     * Throws if record is not in table.
     *
     * @param {(string | number)} id
     * @returns {R}
     * @memberof Table
     */
    get(id: string | number): R;
    /**
     * Gets a single record by it's PK.
     * If not in table, then returns undefined.
     *
     * @param {(string | number)} id
     * @returns {(R | undefined)}
     * @memberof Table
     */
    getOrDefault(id: string | number): R | undefined;
    /**
     * Gets the value of a record by it's PK
     *
     * @param {(string | number)} id
     * @returns {(ValueType<R> | undefined)}
     * @memberof Table
     */
    getValue(id: string | number): ValueType<R> | undefined;
    /**
     * Gets the index used for a given foreign key.
     *
     * @param {string} schemaName
     * @param {string} fkId
     * @returns {string[]}
     * @memberof Table
     */
    getIndex(schemaName: string, fkId: string): string[];
    /**
     * Gets all records in table as wrapped models.
     *
     * @returns {R[]}
     * @memberof Table
     */
    all(): R[];
    /**
     * Gets all values in table.
     *
     * @returns {ValueType<R>[]}
     * @memberof Table
     */
    values(): ValueType<R>[];
    /**
     * Checks whether a record exists in table.
     *
     * @param {(string | number)} id
     * @returns {boolean}
     * @memberof Table
     */
    exists(id: string | number): boolean;
    /**
     * Inserts single or multiple records.
     *
     * @param {Values<R>} data
     * @param {*} [argument] Optional argument passed to normalization.
     * @returns {string[]} Primary keys of inserted records.
     * @memberof Table
     */
    insert(data: Values<R>, argument?: any): string[];
    /**
     * Updates single or multiple records.
     *
     * @param {PartialValues<R>} data
     * @param {*} [argument] Optional argument passed to normalization.
     * @returns {string[]} Primary keys of updated records.
     * @memberof Table
     */
    update(data: PartialValues<R>, argument?: any): string[];
    /**
     * Upserts single or multiple records.
     *
     * @param {PartialValues<R>} data
     * @param {*} [argument] Optional argument passed to normalization.
     * @returns {string[]} Primary keys of upserted records.
     * @memberof Table
     */
    upsert(data: PartialValues<R>, argument?: any): string[];
    /**
     * Deletes single or multiple records by id or object.
     *
     * @param {(string | number | PartialValue<R> | (string | number | PartialValue<R>)[])} data
     * @returns {number} Number of successfully deleted records.
     * @memberof Table
     */
    delete(data: string | number | PartialValue<R> | (string | number | PartialValue<R>)[]): number;
    /**
     * Deletes all records in table.
     *
     * @memberof Table
     */
    deleteAll(): void;
    /**
     * Upserts the table state from another normalized state.
     *
     * @param {TableState<ValueType<R>>} table
     * @memberof Table
     */
    upsertNormalized(table: TableState<ValueType<R>>): void;
    /**
     * Normalizes given data.
     *
     * @param {PartialValues<R>} data
     * @param {NormalizeOptions} [options]
     * @returns {NormalizeContext}
     * @memberof Table
     */
    normalize(data: PartialValues<R>, options?: NormalizeOptions): NormalizeContext;
}
/**
 * Represents a wrapped record belonging to a table.
 *
 * @export
 * @interface TableRecord
 * @template T
 */
export interface TableRecord<T = any> {
    /**
     * Gets the record primary key.
     *
     * @type {string}
     * @memberof TableRecord
     */
    id: string;
    /**
     * Gets the table model which this records belongs.
     *
     * @type {Table}
     * @memberof TableRecord
     */
    table: Table;
    /**
     * Gets or sets the record data value.
     *
     * @type {T}
     * @memberof TableRecord
     */
    value: T;
    /**
     * Update record with new data.
     *
     * @param {Partial<T>} data - Object with partial or full data.
     * @returns {this} Self
     * @memberof TableRecord
     */
    update(data: Partial<T>): this;
    /**
     * Deletes the record from table.
     *
     * @memberof TableRecord
     */
    delete(): void;
}
export interface TableRecordField {
    name: string;
    value: any;
}
export interface TableRecordSet<R extends TableRecord = TableRecord> {
    ids: string[];
    length: number;
    all(): R[];
    values(): ValueType<R>[];
    add(data: Values<R>): string[];
    remove(data: PartialValue<R>): number;
    update(data: PartialValues<R>): string[];
    delete(): void;
}
export interface SchemaNormalizer {
    normalize: (data: any, context: NormalizeContext) => string[];
}
export interface ModelFactory {
    newSchemaNormalizer: (schema: TableSchema) => SchemaNormalizer;
    newSession: (db: DatabaseSchema, state: DatabaseState, options?: SessionOptions) => Session;
    newTableSchema: (db: DatabaseSchema, name: string, schema: TableDefinition) => TableSchema;
    newFieldSchema: (table: TableSchema, name: string, schema: FieldDefinition) => FieldSchema;
    newTableModel: (session: Session, schema: TableSchema, state: TableState) => Table;
    newRecordSetModel: (table: Table, schema: FieldSchema, owner: RecordEntity) => TableRecordSet;
    newRecordFieldModel: (schema: FieldSchema, record: TableRecord) => TableRecordField;
}
export interface RecordFactory {
    newRecordModel: (id: string, table: Table) => TableRecord;
}
/**
 * Defines a database schema
 */
export interface Schema {
    [key: string]: TableDefinition;
}
/**
 * Defines a table schema
 */
export interface TableDefinition {
    [key: string]: FieldDefinition;
}
export interface FieldDefinition {
    /**
     * The field type (Deprecated)
     *
     * @type {FieldType}
     * @memberof FieldDefinition
     * @deprecated
     */
    type?: FieldType;
    pk?: boolean;
    stamp?: boolean;
    /**
     * The name this field references´ in the model. Defaults to schema key.
     *
     * @type {string}
     * @memberof FieldDefinition
     */
    fieldName?: string;
    propName?: string;
    references?: string;
    relationName?: string;
    cascade?: boolean;
    unique?: boolean;
    notNull?: boolean;
    value?: (record: any, context: ComputeContext<any>) => any;
}
export interface ComputeContext<T> {
    schema: FieldSchema;
    record: TableRecord<T>;
}
export interface NormalizeContext {
    table?: Table;
    schema: TableSchema;
    db: DatabaseSchema;
    output: MapOf<TableState>;
    emits: {
        [key: string]: any[];
    };
    normalizePKs: boolean;
    emit(tableName: string, record: any): void;
}
export interface DatabaseSchema {
    tables: TableSchema[];
    options: DatabaseOptions;
    factory: ModelFactory & RecordFactory;
    getRecordNormalizer: (schemaName: string) => RecordNormalizer | undefined;
    getPkGenerator: (schemaName: string) => PkGenerator | undefined;
    getRecordComparer: (schemaName: string) => RecordComparer | undefined;
    getRecordMerger: (schemaName: string) => RecordMerger | undefined;
}
export interface DatabaseOptions {
    onNormalize?: MapOf<RecordNormalizer> | RecordNormalizer;
    onGeneratePK?: MapOf<PkGenerator> | PkGenerator;
    onRecordCompare?: MapOf<RecordComparer> | RecordComparer;
    onRecordMerge?: MapOf<RecordMerger> | RecordMerger;
    cascadeAsDefault?: boolean;
    strict?: boolean;
}
export interface DatabaseCreateOptions extends DatabaseOptions {
    factory?: Partial<ModelFactory>;
    recordModelClass?: RecordClass;
}
export interface SessionOptions {
    readOnly?: boolean;
    tableSchemas?: TableSchema[];
}
export interface NormalizeOptions {
    normalizePKs?: boolean;
    argument?: any;
}
export interface NormalizeContextOptions extends NormalizeOptions {
    table?: Table;
}
export declare type DatabaseState = MapOf<TableState>;
export declare type TableMap = MapOf<Table>;
export interface TableState<T = any> {
    name?: string;
    byId: MapOf<T>;
    ids: string[];
    indexes: MapOf<TableIndex>;
}
export interface TableIndex {
    unique: boolean;
    values: MapOf<string[]>;
}
export interface Session<T extends TableMap = TableMap> {
    db: DatabaseSchema;
    state: DatabaseState;
    tables: T;
    readOnly: boolean;
    upsert(ctx: NormalizeContext): void;
    commit(): DatabaseState;
}
export interface RecordClass {
    new (id: string, table: Table): TableRecord;
}
export interface ExtendedRecord extends TableRecord {
    _fields: {
        [key: string]: any;
    };
}
